////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   : : "    Delphi "
//  * Unit Name : add_tls.dpr
//  * Purpose   :  ,  TLS Callback  Delphi 
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2016.
//  * Version   : 1.0.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//  * Article   : http://alexander-bagel.blogspot.ru/2016/03/early-execution.html
//  ****************************************************************************
//

program add_tls;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  Windows,
  Classes,
  SysUtils,
  Generics.Collections;

procedure ShowHelp;
begin
  Writeln(': AddTls.exe "   "');
  Writeln(' :');
  Writeln('  - 0: TLS Callback  ');
  Writeln('  - 1:      ');
  Writeln('  - 2:    ');
  Writeln('  - 3:   MAP    ');
  Writeln('  - 4:   MAP ');
  Writeln('  - 5:     ');
  Writeln('  - 6:      TLS ');
end;

type
  TSectionData = record
    Index: Integer;
    StartAddr: DWORD;
    SectionName: ShortString;
  end;
  TSectionDataList = TList<TSectionData>;

const
  HardcodeTLS32Offset = 12;

//
//     TLS   -   ,
//   XE  
//      -    
//      :)
// :
// =============================================================================
function GetTlsTableAddr(const FilePath: string): DWORD;
var
  F: TFileStream;
  DOS: TImageDosHeader;
  NT: TImageNtHeaders;
  I: Integer;
  Section: TImageSectionHeader;
begin
  Result := 0;
  //    
  F := TFileStream.Create(FilePath, fmOpenRead or fmShareDenyWrite);
  try
    //  DOS ,    NT
    F.ReadBuffer(DOS, SizeOf(TImageDosHeader));
    F.Position := DOS._lfanew;
    //  NT ,    
    F.ReadBuffer(NT, SizeOf(TImageNtHeaders));
    //     TLS
    for I := 0 to NT.FileHeader.NumberOfSections - 1 do
    begin
      F.ReadBuffer(Section, SizeOf(TImageSectionHeader));
      if PAnsiChar(@Section.Name[0]) = '.tls'  then
      begin
        //  IMAGE_TLS_DIRECTORY,      AddressOfCallback
        Result := Section.PointerToRawData + HardcodeTLS32Offset;
        Break;
      end;
    end;
  finally
    F.Free;
  end;
end;

//        
function GetSectionDataList(const FilePath: string; var Index: Integer): TSectionDataList;
var
  S: TStringList;
  Line: string;
  Section: TSectionData;
begin
  Result := TSectionDataList.Create;
  try
    S := TStringList.Create;
    try
      S.LoadFromFile(FilePath);
      Index := 0;
      Writeln('  ...');
      while Copy(Trim(S[Index]), 1, 5) <> 'Start' do
        Inc(Index);
      Inc(Index);
      while Trim(S[Index]) <> '' do
      begin
        Line := Trim(S[Index]);
        Section.Index := StrToInt(Copy(Line, 1, 4));
        Delete(Line, 1, 5);
        Section.StartAddr := StrToInt('$' + Copy(Line, 1, 8));
        Delete(Line, 1, 19);
        Section.SectionName := ShortString(Trim(Copy(Line, 1, 8)));
        Result.Add(Section);
        Inc(Index);
      end;
      Writeln('  : ', Result.Count);
    finally
      S.Free;
    end;
  except
    //   .   
    on E: Exception do
      Writeln('GetSectionDataList: ' + E.ClassName + ': ' + E.Message);
  end;
end;

// -         tls_callback
//  ( )    ,    
function GetTlsCallbackAddr(const FilePath: string;
  SectionDataList: TSectionDataList; Index: Integer): DWORD;
var
  S: TStringList;
  Line: string;
  SectionIndex, TlsAddr: Integer;
begin
  Result := 0;
  try
    S := TStringList.Create;
    try
      S.LoadFromFile(FilePath);
      Writeln(' tls_callback...');
      repeat
        Line := Trim(S[Index]);
        Inc(Index);
        if Index = S.Count then Break;
      until Pos('.tls_callback', Line) <> 0;
      if Pos('.tls_callback', Line) = 0 then
      begin
        Writeln('     tls_callback  ');
        Exit;
      end;
      SectionIndex := StrToInt(Copy(Line, 1, 4));
      Delete(Line, 1, 5);
      TlsAddr := StrToInt('$' + Copy(Line, 1, 8));
      Writeln('tls_callback , : ', IntToHex(TlsAddr, 8), ', : ', SectionIndex);
      Writeln('   ...');
      for Index := 0 to SectionDataList.Count - 1 do
        if SectionDataList[Index].Index = SectionIndex then
        begin
          Result := SectionDataList[Index].StartAddr + DWORD(TlsAddr);
          Writeln('TLS Callback,    "', SectionDataList[Index].SectionName,
            '",  : ', IntToHex(SectionDataList[Index].StartAddr, 8),
            ', c c: ', IntToHex(Result, 8));
          Break;
        end;
      if Result = 0 then
        Writeln('  tls_callback  ')
    finally
      S.Free;
    end;
  except
    //   .   
    on E: Exception do    
      Writeln('GetTlsCallbackAddr: ' + E.ClassName + ': ' + E.Message);
  end;
end;

//   
function Patch(const FilePath, MapPath: string; TlsTable, CallbackAddr: DWORD): Boolean;
var
  F: TFileStream;
  NewFilePath, BackUpFilePath: string;
  OldCallbackTableAddr: DWORD;
begin
  Result := False;
  try
    NewFilePath := ExtractFilePath(FilePath) + 'tls_aded_' +
      ExtractFileName(FilePath);
    Writeln('  , : ', NewFilePath);
    CopyFile(PChar(FilePath), PChar(NewFilePath), False);
    F := TFileStream.Create(NewFilePath, fmOpenReadWrite);
    try
      Writeln(' ');
      F.Position := TlsTable;
      //      
      F.ReadBuffer(OldCallbackTableAddr, 4);
      //  delphi     SizeOfZeroFill  IMAGE_TLS_DIRECTORY
      //        (   )
      //            
      //      ( 2    32 ,   64  )
      Inc(OldCallbackTableAddr, SizeOf(DWORD) * 2);
      F.Position := TlsTable;
      //      
      F.WriteBuffer(OldCallbackTableAddr, 4);
      Writeln('    , : ', IntToHex(TlsTable, 8),
        ',  : ', IntToHex(OldCallbackTableAddr, 8));
      //          VA   ( RVA)
      //  SizeOfZeroFill  Characteristics     
      F.Position := TlsTable + SizeOf(DWORD) * 3;
      //      
      F.WriteBuffer(CallbackAddr, 4);
      Writeln('  , : ', IntToHex(TlsTable + SizeOf(DWORD) * 3, 8));
      //    ,     
      CallbackAddr := 0;
      F.WriteBuffer(CallbackAddr, 4);
    finally
      F.Free;
    end;
    //   ,   
    Writeln(' ');
    BackUpFilePath := FilePath + '.bak';
    DeleteFile(BackUpFilePath);
    RenameFile(FilePath, BackUpFilePath);
    Writeln(' ');
    RenameFile(NewFilePath, FilePath);
    Writeln('  ');
    Result := True;
  except
    //   .   
    on E: Exception do
    begin
      //       -   
      DeleteFile(NewFilePath);
      RenameFile(BackUpFilePath, FilePath);
      Writeln('Patch: ' + E.ClassName + ': ' + E.Message);
    end;
  end;
end;

var
  MapPath: string;
  TlsTable, CallbackAddr: DWORD;
  SectionDataList: TSectionDataList;
  Index: Integer;
begin
  ExitCode := 0;
  if ParamCount = 0 then
  begin
    ShowHelp;
    ExitCode := 1;
    ExitProcess(ExitCode);
  end;
  if not FileExists(ParamStr(1)) then
  begin
    Writeln('   : ', ParamStr(1));
    ExitCode := 2;
    ExitProcess(ExitCode);
  end;
  TlsTable := GetTlsTableAddr(ParamStr(1));
  if TlsTable = 0 then
  begin
    ExitCode := 6;
    ExitProcess(ExitCode);
  end;
  MapPath := ChangeFileExt(ParamStr(1), '.map');
  if not FileExists(MapPath) then
  begin
    Writeln('   : ', MapPath);
    ExitCode := 3;
    ExitProcess(ExitCode);
  end;
  Index := 0;
  SectionDataList := GetSectionDataList(MapPath, Index);
  try
    if SectionDataList.Count = 0 then
    begin
      Writeln('    ');
      ExitCode := 9;
      ExitProcess(ExitCode);
    end;
    CallbackAddr := GetTlsCallbackAddr(MapPath, SectionDataList, Index);
    if CallbackAddr = 0 then
    begin
      ExitCode := 4;
      ExitProcess(ExitCode);
    end;
    if not Patch(ParamStr(1), MapPath, TlsTable, CallbackAddr) then
      ExitCode := 5;
  finally
    SectionDataList.Free;
  end;
  ExitProcess(ExitCode);
end.
